# generic character controller, controls movement and aiming of a body passed in
# has some simple flocking algorithms, radius of area child node defines neighbor area

extends Area

enum Teams {PLAYER, BANDIT}

const COHESION_STRENGTH = 1.0
const ALIGNMENT_STRENGTH = 0.7
const SEPERATION_STRENGTH = 2.5

export var turn_speed = 500 # deg per second, negative is instant
export(Teams) var team = Teams.PLAYER

var start_team = team
var body = null
var is_frozen = false

export var max_speed = 20.0
export var move_accel = 5.0 # increase velocity by X meters / second
var move_drag = 0.0
export var stop_drag = 0.99

var move_vec = Vector3() # direction to travel
var velocity = Vector3() # direction currently travelling and speed

var neighbor_radius = 0
var flock_point = Vector3()

signal moved
signal rotated

export var verbose = false

func _ready():
	move_drag = move_accel / max_speed
	if body == null:
		body = get_parent()
	neighbor_radius = $CollisionShape.shape.radius

func freeze():
	is_frozen = true
	velocity = Vector3()
	team = -1

func unfreeze():
	is_frozen = false
	team = start_team

func _physics_process(delta):
	if body == null:
		print("body not set")
		return
	if is_frozen:
		return
	
	calculate_movement(delta)

func aim_at(target_pos, t_delta, thing_to_aim=null):
	if thing_to_aim == null:
		thing_to_aim = body
	if is_frozen:
		return
	var our_pos = thing_to_aim.global_transform.origin
	our_pos.y = 0
	target_pos.y = 0
	var to_target = target_pos - our_pos
	var goal_rot = rad2deg(atan2(to_target.x, to_target.z))
	# instantly face target
	if turn_speed < 0: 
		thing_to_aim.rotation_degrees.y = goal_rot
	# otherwise slowly rotate to target
	elif abs(angle_difference(goal_rot, thing_to_aim.rotation_degrees.y)) < turn_speed * t_delta:
		# prevent overshooting goal angle
		thing_to_aim.rotation_degrees.y = goal_rot
	else:
		#calc which direction to turn
		var right = thing_to_aim.global_transform.basis.x
		var turn_dir = 1
		if right.dot(to_target) < 0:
			turn_dir = -1
		thing_to_aim.rotation_degrees.y += turn_speed * turn_dir * t_delta

	emit_signal("rotated", t_delta)

func angle_difference(from, to):
    return fposmod(to - from + 180,  180 * 2) - 180

func set_move_vec(new_vec):
	move_vec = Vector3(new_vec.x, 0, new_vec.z).normalized()

func calculate_movement(delta):
	var h = body.global_transform.origin.y
	
	var dir_of_travel = calc_move_vec_with_flocking()
	var drag = move_drag
	if dir_of_travel.dot(velocity) < 0 or dir_of_travel == Vector3():
		drag = stop_drag
	velocity += dir_of_travel * move_accel - velocity * drag
	velocity = body.move_and_slide(velocity, Vector3())
	if velocity.length() < 0.05:
		velocity = Vector3()
	#if verbose:
	#	print(velocity.length())
	
	# lock height, don't want to move up or down at all
	body.global_transform.origin.y = h
	emit_signal("moved", delta)
	

# use flocking algorithm to figure out what direction to move, very roughly implemented lol
func calc_move_vec_with_flocking():
	# flocking calculations
	var our_pos = global_transform.origin
	our_pos.y = 0
	flock_point = our_pos
	var neighboring_units = []
	for n in get_overlapping_areas():
		if "team" in n and n.team == team:
			neighboring_units.append(n)
	var num_of_neighbors = neighboring_units.size()
	if num_of_neighbors == 0:
		return move_vec
	
	var cohesion_vec = Vector3()
	var alignment_vec = Vector3()
	
	for n in neighboring_units:
		var n_pos = n.global_transform.origin
		n_pos.y = 0
		var dis = our_pos.distance_to(n_pos)
		if dis > neighbor_radius:
			continue
		var vec_to_n = ((n_pos - our_pos) / dis)
		flock_point += n_pos
		
		var dist_norm = clamp(dis / neighbor_radius, 0.0, 1.0) # 0 when next to us, 1 when max distance
		
		cohesion_vec += vec_to_n * dist_norm - vec_to_n * (1.0 - dist_norm) * SEPERATION_STRENGTH
		if verbose:
			print(dis, " ", dist_norm, " ", neighbor_radius, cohesion_vec)
		alignment_vec += n.move_vec * (1.0 - dist_norm)
	
	alignment_vec /= num_of_neighbors
	alignment_vec *= ALIGNMENT_STRENGTH
	flock_point /= (num_of_neighbors + 1)
	
	return move_vec + cohesion_vec.normalized() * 0.04 + alignment_vec * 1.1

# used by player
func update_stats(new_stats: Dictionary):
	max_speed = new_stats.speed
	move_accel = max_speed / 4.0
	move_drag = move_accel / max_speed