extends KinematicBody

onready var character_controller = $CharacterController
onready var sight_area = $SightArea
onready var health_manager = $HealthManager
onready var anim = $Graphics/AnimationTree
onready var weapon = $"Graphics/Scene Root/Armature/Skeleton/Hand".get_child(0)

enum States {
	REST, 
	CHASE, 
	DEAD
}
var cur_state = States.REST

signal spotted_enemy

const MIN_DIS_TO_PLAYER = 2.0
const HEARING_RANGE = 5.0
const MIN_DIS_TO_PATH_NODE = 0.1
const MAX_TIME_ENEMY_OUT_OF_RANGE = 3.0
const RESET_PATH_WHEN_ENEMY_THIS_FAR_FROM_END = 2.0

var cur_zigzag_time = 0.0
var cur_time_enemy_out_of_range = 0.0
var enemy_in_sight_area = false

export var zigzag_speed = 0.6 # time to perform one complete zigzag
export var sight_arc = 0.6 # 0.0 - 1.0

var enemy = null
var start_position = Vector3()
var start_facing = Vector3()
var path = []

var override = false # override state switching behaviours
var alert_when_reached_point = null
signal reached_point

func _ready():
	enemy = get_tree().get_nodes_in_group("player")[0]
	$Container.get_node("CollisionShape").disabled = true
	health_manager.connect("died", self, "switch_to_dead_state")
	sight_area.connect("body_entered", self, "something_entered_sight_range")
	sight_area.connect("body_exited", self, "something_exited_sight_range")
	
	start_position = global_transform.origin
	start_facing = start_position + global_transform.basis.z
	weapon.fire_point = $FirePoint
	weapon.set_objects_to_ignore([self])
	weapon.connect("reloading", anim, "reload")
	# godot is dumb and doesn't let me connect this in the editor so I have to do it from code
	$HealthManager.connect("damaged", $"Graphics/Scene Root/Armature/Skeleton/Bandit", "update_bleed_amount")

var t_delta = 0
func _process(delta):
	t_delta = delta
	if cur_state == States.REST:
		execute_rest_state()
	elif cur_state == States.CHASE:
		execute_chase_state()

#REST STATE
# - if we're not at starting position, return there
# - else stand still
# - if enemy is spotted or we are shot at, switch to CHASE STATE
func switch_to_rest_state():
	if cur_state == States.DEAD:
		return
	cur_state = States.REST
	path = get_parent().get_simple_path(global_transform.origin, start_position)

func execute_rest_state():
	if enemy == null:
		return
	var move_vec = get_move_vec_from_path()
	character_controller.set_move_vec(move_vec)
	
	if enemy_in_sight_area and (can_see_enemy() or can_hear_enemy()):
		switch_to_chase_state()
	if move_vec != Vector3():
		character_controller.aim_at(global_transform.origin + move_vec, t_delta)
	else:
		character_controller.aim_at(start_facing, t_delta)
	anim.update_body(move_vec, global_transform.basis.z)

# CHASE STATE
# - follow path to enemy
# - if enemy moves beyond certion distance from path end, recalculate path
# - if can see enemy, shoot at enemy
# - if enemy sees us, zigzag
# - if enemy out of sight range, start counting
# - - if MAX_TIME_ENEMY_OUT_OF_RANGE seconds pass and still can't see enemy, switch to REST STATE
func switch_to_chase_state():
	if cur_state == States.DEAD:
		return
	cur_state = States.CHASE
	cur_zigzag_time = 0.0
	cur_time_enemy_out_of_range = 0.0
	emit_signal("spotted_enemy", enemy)

func execute_chase_state():
	if enemy == null:
		return
	
	# movement logic
	var our_pos = global_transform.origin
	var enemy_pos = enemy.global_transform.origin
	if should_generate_new_chase_path():
		path = get_parent().get_simple_path(global_transform.origin, enemy_pos)
	var move_vec = get_move_vec_from_path()
	var dis_to_enemy = (enemy_pos - our_pos).length()
	if dis_to_enemy < MIN_DIS_TO_PLAYER:
		move_vec = Vector3()
	
	# zigzag logic
	var dir_from_enemy = (global_transform.origin - enemy_pos).normalized()
	var dir_enemy_facing = enemy.global_transform.basis.z
	var dir_to_enemy = -dir_from_enemy
	# check if enemy is aiming at us
	if has_line_of_sight_on_enemy() and dir_from_enemy.dot(dir_enemy_facing) > 0.9:
		var left_or_right = sign(cos(cur_zigzag_time * PI / zigzag_speed))
		var zigzag_vec = dir_from_enemy.rotated(Vector3.UP, PI / 2.0) * left_or_right
		#zig zag less the more perpendicular from enemy we are travelling 
		zigzag_vec *= abs(dir_to_enemy.dot(move_vec))
		move_vec = (move_vec + zigzag_vec).normalized()
		cur_zigzag_time += t_delta
	
	# aiming logic
	if has_line_of_sight_on_enemy() or dis_to_enemy < 10:
		character_controller.aim_at(enemy_pos, t_delta)
	else:
		character_controller.aim_at(our_pos + move_vec, t_delta)
	character_controller.set_move_vec(move_vec)

	# perform attack logic
	if can_see_enemy() and aiming_at_player():
		weapon.auto_attack()
		#TODO attach reloading signal to animation
	
	anim.update_body(character_controller.velocity, global_transform.basis.z)
	
	if !enemy_in_sight_area:
		cur_time_enemy_out_of_range += t_delta
		if !override and cur_time_enemy_out_of_range > MAX_TIME_ENEMY_OUT_OF_RANGE:
			switch_to_rest_state()

func switch_to_dead_state():
	cur_state = States.DEAD
	character_controller.freeze()
	$CollisionShape.disabled = true
	$Container.get_node("CollisionShape").disabled = false
	anim.update_dead()

func should_generate_new_chase_path():
	if override:
		return false
	var our_pos = global_transform.origin
	var enemy_pos = enemy.global_transform.origin
	# are we too close to enemy
	if our_pos.distance_squared_to(enemy_pos) < pow(MIN_DIS_TO_PATH_NODE, 2):
		return false
	# has enemy moved too far from last point of the path
	if path.size() == 0 or path[path.size()-1].distance_squared_to(enemy_pos) > pow(RESET_PATH_WHEN_ENEMY_THIS_FAR_FROM_END, 2):
		return true
	return false

func something_entered_sight_range(coll):
	if "Bullet" in coll.name and cur_state == States.REST and coll.shot_by == "Player": # we got shot at
		switch_to_chase_state()
	if "Player" in coll.name:
		enemy_in_sight_area = true
		cur_time_enemy_out_of_range = 0.0

func something_exited_sight_range(coll):
	if "Player" in coll.name:
		enemy_in_sight_area = false

func can_see_enemy():
	var enemy_pos = enemy.global_transform.origin
	var our_pos = global_transform.origin
	var dir_to_enemy = (enemy_pos - our_pos).normalized()
	var in_sight_arc = dir_to_enemy.dot(global_transform.basis.z) > 1.0 - sight_arc
	if !in_sight_arc:
		return false
	return has_line_of_sight_on_enemy()

func aiming_at_player():
	var enemy_pos = enemy.global_transform.origin
	var our_pos = global_transform.origin
	var dir_to_enemy = (enemy_pos - our_pos).normalized()
	return dir_to_enemy.dot(global_transform.basis.z) > 0.9

func has_line_of_sight_on_enemy():
	var enemy_pos = enemy.global_transform.origin
	var our_pos = global_transform.origin
	var space_state = get_world().direct_space_state
	var result = space_state.intersect_ray(our_pos, enemy_pos, [self], collision_mask)
	if !result:
		return false
	return result.collider == enemy

func can_hear_enemy():
	return global_transform.origin.distance_squared_to(enemy.global_transform.origin) < HEARING_RANGE * HEARING_RANGE

func get_move_vec_from_path():
	var move_vec = Vector3()
	if path.size() > 0: 
		var dir_to_path_node = path[0] - global_transform.origin
		dir_to_path_node.y = 0
		var dis_to_path_node = dir_to_path_node.length()
		if dis_to_path_node < MIN_DIS_TO_PATH_NODE: # or character_controller.flock_point.distance_squared_to(path[0]) < pow(MIN_DIS_TO_PATH_NODE, 2):
			path.remove(0)
		else:
			move_vec = dir_to_path_node / dis_to_path_node
		if alert_when_reached_point != null:
			if global_transform.origin.distance_squared_to(alert_when_reached_point) < MIN_DIS_TO_PATH_NODE * MIN_DIS_TO_PATH_NODE:
				emit_signal("reached_point", self)
	return move_vec

func damage(amnt, dir):
	if cur_state == States.REST:
		switch_to_chase_state()
	health_manager.damage(amnt, dir)