extends KinematicBody

onready var combat_manager = $CombatManager
onready var character_controller = $CharacterController
onready var health_manager = $CombatManager/HealthManager
onready var inventory = $Inventory
onready var camera = $CameraBase/CamPosOffset/Camera
onready var anim = $Graphics/AnimationTree

enum States {
	NORMAL, 
	INVENTORY, 
	DEAD
}
var cur_state = States.NORMAL
var t_delta = 0 # I store delta here so I don't have to pass it into every method
var cam_basis
var cam_start_rotation = 0

func _ready():
	cam_basis = $CameraBase.global_transform.basis
	cam_start_rotation = cam_basis.get_euler().y
	character_controller.body = self
	character_controller.turn_speed = -1
	#$CombatManager/HealthManager.connect("died", self, "set_state_dead")
	combat_manager.fire_point = $FirePoint
	exit_all_states()
	set_state_normal()
	_load()

func _process(delta):
	t_delta = delta
	if Input.is_action_just_pressed("exit"):
		_save()
		get_tree().quit()
	
	if cur_state == States.DEAD:
		return
	if cur_state == States.NORMAL:
		execute_normal_state()
	elif cur_state == States.INVENTORY:
		execute_inventory_state()
	$CameraBase.global_transform.basis = cam_basis

func set_state_normal():
	exit_all_states()
	cur_state = States.NORMAL
	character_controller.unfreeze()
	combat_manager.dont_update_cursor = false
	combat_manager.update_crosshair()
	
func set_state_inventory():
	exit_all_states()
	cur_state = States.INVENTORY
	update_inventory_scale()
	inventory.open()

func set_state_dead():
	exit_all_states()
	cur_state = States.DEAD

func exit_all_states():
	character_controller.freeze()
	combat_manager.dont_update_cursor = true
	#combat_manager.stop_all()
	anim.update_body(Vector3(), global_transform.basis.z, false)
	inventory.close()

func execute_normal_state():
	if Input.is_action_just_pressed("toggle_inventory"):
		set_state_inventory()
		return
	var move_vec = Vector3()
	if Input.is_action_pressed("move_up"):
		move_vec.z += 1
	if Input.is_action_pressed("move_down"):
		move_vec.z -= 1
	if Input.is_action_pressed("move_right"):
		move_vec.x -= 1
	if Input.is_action_pressed("move_left"):
		move_vec.x += 1
	move_vec = move_vec.rotated(Vector3(0, 1, 0), cam_start_rotation)
	character_controller.set_move_vec(move_vec)
	var moving = move_vec != Vector3()
		
	if Input.is_action_just_pressed("hotkey_0"):
		combat_manager.switch_weapon(0)
	if Input.is_action_just_pressed("hotkey_1"):
		combat_manager.switch_weapon(1)
	if Input.is_action_just_pressed("hotkey_2"):
		combat_manager.switch_weapon(2)
	if Input.is_action_just_pressed("hotkey_3"):
		combat_manager.switch_weapon(3)
	
	if Input.is_action_just_pressed("attack"):
		combat_manager.single_attack()
	if Input.is_action_pressed("attack"):
		combat_manager.auto_attack()
	if Input.is_action_just_pressed("reload"):
		combat_manager.reload()
	
	if not moving and Input.is_action_pressed("eat") and not combat_manager.at_full_health():
		combat_manager.eat()
		var meat = health_manager.get_nearest_meat()
		if meat != null and combat_manager.is_eating():
			character_controller.aim_at(meat.global_transform.origin, t_delta)
	else:
		combat_manager.stop_eating()
	
	if not combat_manager.is_eating():
		character_controller.aim_at(camera.get_point_under_mouse(), t_delta)
	
	anim.update_body(character_controller.velocity, global_transform.basis.z, combat_manager.is_eating())
	
func execute_inventory_state():
	if Input.is_action_just_pressed("toggle_inventory"):
		set_state_normal()
		return
	
	var cursor_pos = get_viewport().get_mouse_position()
	# some code to let me scale the inventory arbitrarily
	cursor_pos -= inventory.offset
	cursor_pos /= inventory.scale
	
	if Input.is_action_just_pressed("inventory_interact"):
		inventory.grab_item(cursor_pos)
	if Input.is_action_pressed("inventory_interact"):
		inventory.drag_item(cursor_pos)
	if Input.is_action_just_released("inventory_interact"):
		inventory.release_item()
	if Input.is_action_just_pressed("rotate"):
		inventory.rotate_item(cursor_pos)
	inventory.display_item_info(cursor_pos)

func update_inventory_scale():
	var vp_size = get_viewport().size
	inventory.offset = vp_size / 2 * (Vector2(1, 1) - inventory.scale)

func damage(amnt, dir):
	combat_manager.damage(amnt, dir)

# TODO make this arbitrary, player should be in persist group as well
func _save():
	var save_game = File.new()
	save_game.open("user://savegame.save", File.WRITE)
	var save_data = {}
	save_data["inventory"] = inventory._save()
	save_data["last_hotkey"] = combat_manager.cur_weapon_slot
	for node in get_tree().get_nodes_in_group("persist"):
		save_data[str(node)] = node._save()
	save_game.store_line(to_json(save_data))
	save_game.close()
	#TODO save instanced objects e.g. corpses, blood splatters

func _load():
	var save_game = File.new()
	if not save_game.file_exists("user://savegame.save"):
		return
	save_game.open("user://savegame.save", File.READ)
	var save_data = parse_json(save_game.get_line())
	save_game.close()
	if save_data == null:
		return
	if "inventory" in save_data:
		inventory._load(save_data["inventory"])
	if "last_hotkey" in save_data:
		combat_manager.switch_weapon(int(save_data["last_hotkey"]))
	
	for node in get_tree().get_nodes_in_group("persist"):
		if str(node) in save_data:
			node._load(save_data[str(node)])
	#TODO instance the rest e.g. corpses, blood splatters

func clear_save():
	var save_game = File.new()
	save_game.open("user://savegame.save", File.WRITE)
	save_game.store_line(to_json({}))
	save_game.close()


func add_item():
	pass # Replace with function body.
